import logging
import os.path
import shutil
import subprocess

from .ArchiveCache import ArchiveCache
from .DesktopFinder import DesktopFinder
from .IconFinder import IconFinder, ICON_EXTS
from .utils import load_deb_file, file_charset

# Extra data to app for every desktop entry
_METADATA = "\nX-AppInstall-Package={0}\nX-AppInstall-Section={1}\n"


def extract_metadata(archive_url, suite, arches, out_dir, memory):
    """
    This method extracts icons and desktop entries for applications from
    a debian archive. The archive must conform to standards explained in
    <http://www.debian.org/doc/FAQ/ch-ftparchives.html>. It must also
    generate the (optional) Contents-<ARCH>.gz files.

    Application desktop entries will extract in `out_dir/desktop`. Icons
    will extract in `out_dir/icons`.

    @archive_url: uri to an archive with a 'dists' path
    @suite: a dist series accessible in 'archive_url/dists/'
    @arches: a list of package architectures to search in
    @out_dir: a directory where extracted metadata should output
    @memory: the maximum number of megabytes to load into memory
    """
    # make the extraction paths
    apps_dir = os.path.join(out_dir, "desktop")
    icons_dir = os.path.join(out_dir, "icons")
    if os.path.exists(apps_dir):
        shutil.rmtree(apps_dir)
    if os.path.exists(icons_dir):
        shutil.rmtree(icons_dir)
    os.makedirs(apps_dir)
    os.makedirs(icons_dir)

    # set up the archive cache: it provides an index for searching paths and
    # packages in an archive, as well as obtaining their debs.
    search_prefixes = (DesktopFinder.prefixes + IconFinder.prefixes)
    release = os.path.join(archive_url, "dists", suite, "Release")
    cache = ArchiveCache(release, search_prefixes, memory)
    assert "Description" in cache.release and "Date" in cache.release
    logging.info("Extracting app data from {0}, last modified {1}".format(
                 cache.release["Description"], cache.release["Date"]))

    # set up search objects. we should not repeat extraction for the same app
    # packages on different architectures.
    desktop_finder = DesktopFinder()
    apps_already_found = set()

    for arch in arches:
        # get all desktop entry paths and packages from the archive cache.
        logging.info("Now looking at {0} architecture".format(arch))
        cache.load_arch(arch)
        desktop_entries = desktop_finder.search(cache.contents_index)
        # filter out those apps found in previous architectures.
        new_entries = set(desktop_entries).difference(apps_already_found)
        apps_already_found = set(new_entries).union(apps_already_found)
        logging.info("{0} new applications found in {1}".format(
                     len(new_entries), arch))
        # extract the desktop entry and its metadata to the output paths.
        for package, entry in new_entries:
            logging.debug("Now reading at {0} in {1}".format(entry, package))
            _extract_app_data(entry, package, cache, apps_dir, icons_dir)
    logging.info("Extraction complete. {0} applications processed".format(
                 len(apps_already_found)))


def _extract_app_data(entry, package, cache, apps_dir, icons_dir):
    # secure the latest deb for a package from the archive cache
    deb = cache[package]
    if deb and os.path.exists(deb):
        logging.debug("{0} successfully cached at {1}".format(package, deb))
    else:
        logging.warning("Could not cache deb for {0}".format(package))
        return
    # get the desktop entry itself. if load_deb_file returns another path
    # rather than a file object, try to search for that path in the cache.
    desktop_file = load_deb_file(entry, deb)
    if isinstance(desktop_file, str):
        linked_package = cache.contents_index.search_exact(desktop_file)
        # FIXME: decide how many symlinked paths we should follow before
        # giving up on finding the file.
        if not linked_package:
            return
        linked_deb = cache[linked_package]
        desktop_file = load_deb_file(desktop_file, linked_deb)
    if not desktop_file or isinstance(desktop_file, str):
        logging.warning("Could not extract {0} from {1}".format(entry, deb))
        return
    desktop_file_data = desktop_file.read()
    desktop_file.close()
    # read the desktop file into a sane encoding
    # TODO: decide how to handle error checking for truly insane encodings
    try:
        charset = file_charset(desktop_file.name)
        desktop_lines = desktop_file_data.decode(charset).split("\n")
    except Exception:
        logging.warn("Couldn't read {0}, trying latin-1".format(entry))
        desktop_lines = desktop_file_data.decode("latin-1").split("\n")
    desktop_name = os.path.basename(entry)
    desktop_path = os.path.join(apps_dir, desktop_name)
    # prevent overwriting of data for same-desktop-named apps
    if os.path.exists(desktop_path):
        desktop_path = os.path.join(apps_dir, entry.replace("/", "_"))
    # finally, rewrite the desktop entry with additional metadata, and
    # extract any icons requested to the icon output direcotry.
    with open(desktop_path, "wb") as f:
        for line in desktop_lines:
            if line.startswith("Icon="):
                icon_name = line.split("=", 1)[1].strip()
                if icon_name.startswith("/"):
                    out_name = icon_name.replace("/", "_")
                else:
                    # only take the file name for relative paths
                    out_name = os.path.basename(icon_name)
                if out_name:
                    out_path = os.path.join(icons_dir, out_name)
                    _extract_icon(icon_name, out_path, cache)
                    # strip off the extension in the desktop file Icon line
                    if any([out_name.endswith(ext) for ext in ICON_EXTS]):
                        out_name = os.path.splitext(out_name)[0]
                    line = line.replace(icon_name, out_name)
            elif line.startswith("[Desktop Entry]"):
                # write metadata now
                section = cache.get_component(package)
                line = "".join([line, _METADATA.format(package, section)])
            f.write(line.encode() + b"\n")
    logging.debug("{0} successfully extracted from {1}".format(entry, deb))


def _extract_icon(search_term, out_path, cache):
    # ask IconFinder for an icon and its package
    finder = IconFinder()
    icon_path = finder.search(search_term, cache)
    if icon_path:
        # add extension to extracted icon if needed
        if not os.path.splitext(out_path)[1]:
            out_path += os.path.splitext(icon_path)[1]
        os.rename(icon_path, out_path)
