import gzip
import hashlib
import logging
import os.path
import re
import tempfile
import urllib


class ContentsIndex(object):
    """
    This class caches the file paths in a deb archive--as described by
    that archive's Contents-<ARCH>.gz--to make search of them efficient.
    It can also index subsets of the archives (eg all paths beginning
    with /usr/share/icons/) for quicker informed searches.

    Indexes are created at instantiation via the prefixes_to_index
    argument. They can then be searched simply by using the `prefix`
    argument in search methods.

    In the event of ContentsIndex runs out of memory to allocate, it
    will prioritize the indexes in order, and write all excess data to
    cache files.
    """

    # From a design standpoint, this class merely wraps around several
    # _Index(prefix_to_cache) instances for searching Contents files.

    def __init__(self, contents_file, prefixes_to_index=[], memory=500):
        """
        @contents_file: The path to a Contents-<ARCH>.gz file.
        @prefixes_to_index: A list of path prefixes (eg /usr/share is a
            prefix of /usr/share/*) to separately index and make
            searchable for better speeds.
        @memory: The maximum amount of data in the contents_file (in MB)
            that the cache will load into memory.
        """
        self._contents_file = contents_file
        self._memory, self._max_memory = 0, memory * 1024 * 1024
        self._num_lines = 0
        self._create_indexes(prefixes_to_index)
        self._load_cache()

    def search(self, regex, prefix=""):
        """
        Returns a list of (package name, path) pairs, where "path" is a
        file location in the archive that the search term matches, and
        "package name" is the name of a package that contains it.

        @regex: A string regular expression to filter paths with.
        @prefix: A string path from prefix_indexes. Only paths starting
            with the prefix will be searched. Speeds up queries.
        """
        prefix = prefix.strip("/")
        if prefix:
            index = self._indexes[prefix]
            logging.debug("Searching cache for all files '%s' with prefix %s" %
                         (regex, prefix))
        else:
            index = self._contents_index
            logging.debug("Searching cache for all files '%s'" % regex)
        regex = regex.encode()
        results = index.search(regex)
        for p in self._prefixes:
            if p.startswith(prefix) and p != prefix:
                child_index = self._indexes[p]
                results.extend(child_index.search(regex))
        return results

    def search_exact(self, path):
        """
        Returns the name of a package that contains the exact `path`
        location, or None if `path` does not exist in the archive.
        """
        path = path.strip("/")
        logging.debug("Searching cache for '%s'" % path)
        path = path.encode()
        prefix = self._find_index(path)
        if prefix:
            return self._indexes[prefix].search_exact(path)
        return self._contents_index.search_exact(path)

    def _create_indexes(self, prefixes):
        self._indexes = {}
        self._prefixes = tuple([p.strip("/") for p in prefixes])
        # indexes for user-defined paths like '/usr/share/icons'
        for prefix in self._prefixes:
            self._indexes[prefix] = _Index()
        # an index for lines of all other paths
        self._contents_index = _Index()

    def _load_cache(self):
        logging.info("Loading archive paths from %s into cache" %
                     os.path.basename(self._contents_file))
        contents = gzip.open(self._contents_file, "rb")
        while not re.match(b"^FILE\W+LOCATION", contents.readline()):
            continue
        for line in contents:
            self._memory += self._place_in_index(line)
            if self._memory > self._max_memory:
                self._reduce_memory_use()
            self._num_lines += 1
        contents.close()
        logging.info("Done loading cache. %iMB in active memory." %
                     (self._memory / 1048576))

    def _place_in_index(self, line):
        # put a line in its right index and return the amount of memory it took
        path = line.split()[0]
        prefix = self._find_index(path)
        if prefix:
            return self._indexes[prefix].add_line(line)
        else:
            return self._contents_index.add_line(line)

    def _find_index(self, path):
        # get the longest index path prefixing the given path
        candidate_index = b""
        for prefix in self._prefixes:
            p = prefix.encode()
            if path.startswith(p) and len(p) > len(candidate_index):
                candidate_index = prefix
        return candidate_index

    def _reduce_memory_use(self):
        if self._contents_index.bytes_loaded:
            logging.info("* Max cache load exceeded. Removing non-indexed"
                         "paths from memory")
            self._contents_index.flush_memory()
            return
        for prefix in reversed(self._prefixes):
            if self._indexes[prefix].bytes_loaded:
                logging.info("Max cache load exceeded. Removing '%s' index "
                             "from memory" % prefix)
                self._indexes[prefix].flush_memory()
                return

    def __len__(self, prefix=""):
        """
        Returns the number of paths searchable in this cache. Assign
        `prefix` to count the paths in a particular index.
        """
        if not prefix:
            return self._num_lines
        else:
            # the length of the prefix and its children (eg /usr -> /usr/share)
            prefix = prefix.strip("/")
            num_lines = self._indexes[prefix].num_lines
            for p in self._prefixes:
                if p.startswith(prefix + "/"):
                    num_lines += self._indexes[p].num_lines
            return num_lines

    def __contains__(self, prefix):
        return prefix.strip("/") in self._prefixes

    contents_file = property(lambda self: self._contents_file, doc="""
        The Contents-<ARCH>.gz file this object indexes""")

    prefixes = property(lambda self: self._prefixes, doc="""
        The path prefixes that the cache has indexed for searching""")


class _Index:
    # a helper implementation class for indexes: cached lines all beginning
    # with the same prefix
    def __init__(self):
        self.bytes_loaded = 0
        self.num_lines = 0
        self._file = tempfile.TemporaryFile()
        self._lines = []
        self._using_memory = True

    def add_line(self, line):
        # add a line to the cache file, and memory if possible
        self.num_lines += 1
        self._file.write(line)
        if self._using_memory:
            to_store = line.strip()
            self._lines.append(to_store)
            bytes_added = len(to_store)
            self.bytes_loaded += bytes_added
            return bytes_added
        return 0

    def flush_memory(self):
        # stop caching lines in memory
        self._lines = None
        self.bytes_loaded = 0
        self._using_memory = False

    def search_exact(self, path):
        if self._using_memory:
            line = _search_lines_exact(path, self._lines)
        else:
            line = _search_file_exact(path, self._file)
        if not line:
            return None
        # the right-most part of a Contents file is a package for the path
        parts = line.decode().rsplit("/", 1)
        if len(parts) != 2:
            logging.warning("Invalid contents line: %s" % line)
            return None
        return parts[1].strip()

    def search(self, regex):
        lines = _search_pattern(regex, self._file, self._lines)
        results = []
        for line in lines:
            parts = line.decode().split()
            if len(parts) != 2:
                logging.warning("Invalid contents line: %s" % line)
                continue
            package_data = parts[1].split(",")
            for data in package_data:
                package_data = data.rsplit("/", 1)
                if len(package_data) != 2:
                    logging.warning("Invalid contents line: %s" % line)
                    continue
                path, package_name = parts[0], package_data[1]
                results.append((package_name, path))
        return results

    def __del__(self):
        self._file.close()


# Utility functions for quickly searching files according to the
# following line format used by Contents-<ARCH>:
#
#    absolute/path/to/file          section/cat/pkgname

def _search_file_exact(search_term, file):
    # Searches the file at `file` for a line starting with
    # `search_term`. Assumes lines are in alphabetical order. Returns
    # None if there is no match.
    file.seek(0, 2)
    low, high = 0, file.tell()

    # perform a binary search on the file
    while high > low:
        mid = (low + high) // 2
        file.seek(mid)

        # read current line from the beginning
        while file.tell() != 0 and file.read(1) != b'\n':
            file.seek(-2, os.SEEK_CUR)
        line_begin = file.tell()
        line = file.readline()
        line_end = file.tell()

        first_term = line.split()[0]
        if search_term.startswith(first_term):
            # return if first term matches search term
            if len(search_term) == len(first_term):
                return line
            # or if the first term looks like a directory (symlink)
            elif (len(search_term) > len(first_term) and
                  chr(search_term[len(first_term)]) == "/"):
                return line
        # otherwise continue the binary search
        if search_term < first_term:
            high = line_begin
        elif search_term > first_term:
            low = line_end
        else:
            return line


def _search_lines_exact(search_term, lines):
    # Searches `lines` for a line starting with `search_term`. Assumes
    # lines are in alphabetical order. Returns None if there is no match.
    low, high = 0, len(lines)
    while high > low:
        mid = (low + high) // 2
        first_term = lines[mid].split()[0]
        if search_term.startswith(first_term):
            # return if first term matches search term
            if len(search_term) == len(first_term):
                return lines[mid]
            # or if the first term looks like a directory (symlink)
            elif (len(search_term) > len(first_term) and
                  chr(search_term[len(first_term)]) == "/"):
                return lines[mid]
        # otherwise continue the binary search
        if search_term < first_term:
            high = mid
        elif search_term > first_term:
            low = mid + 1
        else:
            return lines[mid]


def _search_pattern(search_pattern, search_file, lines=None):
    # Searches the file at `file` or the `line_list` for all lines where
    # the first term (whitespace delimited) matches the `search_pattern`
    # regex string. Use `lines` if they are already loaded.
    pattern = re.compile(search_pattern.replace(b'+', b'\+'))
    results = []
    if lines == None:
        lines = search_file
        lines.seek(0)
    for line in lines:
        # translate file bytes data to unicode
        first_term = line.split()[0]
        if pattern.match(first_term):
            results.append(line)
    return tuple(results)
