import logging
import os.path
import subprocess
import tempfile

from ..utils import load_deb_file

# An ordered list of places to search for icons, chosen to
# minimize search an maximize icons quality.
_SEARCH_POLICY = (
  "usr/share/icons/hicolor/scalable",
  "usr/share/icons/hicolor/128x128",
  "usr/share/icons/hicolor/256x256",
  "usr/share/icons/hicolor/64x64",
  "usr/share/icons/hicolor/48x48",
  "usr/share/pixmaps",
  "usr/share/icons/hicolor/32x32",
  "usr/share/icons/hicolor/22x22",
  "usr/share/icons/hicolor/16x16",
  "usr/share/icons/Humanity",
  # FIXME: Remove when apt-ftparchive
  # has better symlink support
  "usr/lib/GNUstep",
  "usr/share/icons",
  "usr/share")

# Icon formats to accept for applications
ICON_MIMES = (
  "image/gif", "image/jpeg", "image/png", "image/svg+xml",
  "image/tiff", "image/x-ico", "text/x-c")

ICON_EXTS = (".svg", ".png", ".xpm", ".tiff", ".gif",
             ".ico", ".jpg", ".jpeg", ".svgz")

_ICON_REGEX = ".*/{0}({exts}|)$".format("{0}", exts="|".join(ICON_EXTS))


class IconFinder:
    """
    Searches for application icons according to a search policy. Queries
    require a ArchiveCache where all paths in 'IconFinder.prefixes' are
    indexed.
    """

    def search(self, icon_name, cache):
        """
        Searches the archive for the best icon with a given name.

        @icon_name: A valid Icon string for an XDG Desktop Entry.
        @cache: A ArchiveCache with this finder's prefixes indexed.
        @returns '/path/to/icon' or None
        """
        if icon_name.startswith("/"):
            # match an absolute path
            package = cache.contents_index.search_exact(icon_name)
            if package:
                return _extract_icon(icon_name, package, cache)
            # if the path has no extention, search all valid ones.
            # this is actually fairly cheap.
            if os.path.splitext(icon_name)[1] not in ICON_EXTS:
                for ext in ICON_EXTS:
                    returned = self.search(icon_name + ext, cache)
                    if returned:
                        return returned
        else:
            # match any path ending with the name and an extension
            icon_name = self._extensionless(icon_name)
            pattern = _ICON_REGEX.format(icon_name)
            for prefix in self.prefixes:
                results = cache.contents_index.search(pattern, prefix)
                if results:
                    package, icon_name = results[0]
                    return _extract_icon(icon_name, package, cache)

    def _extensionless(self, path):
        # utility function to intelligently strip icon extensions
        if any([path.endswith(ext) for ext in ICON_EXTS]):
            return os.path.splitext(path)[0]
        return path

    prefixes = _SEARCH_POLICY


def _extract_icon(icon, package, cache):
    logging.debug("Found icon and package {0}, {1}".format(icon, package))
    deb = cache[package]
    if not deb:
        logging.warning("Could not find '{0}' for {1}".format(package, icon))
        return
    # load content from the deb and save it
    icon_file = load_deb_file(icon, deb)
    if isinstance(icon_file, str):
        linked_package = cache.contents_index.search_exact(icon_file)
        if not linked_package:
            return
        linked_deb = cache[linked_package]
        icon_file = load_deb_file(icon_file, linked_deb)
    if not icon_file or isinstance(icon_file, str):
        logging.warning("Bailing from " + str(icon_file) + " extraction")
        return
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(icon_file.read())
    icon_file.close()
    # validate it is an icon
    filetype = subprocess.check_output(["file", "-zbi", f.name]).decode()
    mimetype = filetype.split(";")[0]
    if mimetype not in ICON_MIMES:
        logging.warning("'{0}' is not an icon".format(icon))
        os.remove(f.name)
        return
    logging.debug("Got icon {0}: {1}".format(icon, f.name))
    # give extension to the file
    extension = os.path.splitext(icon)[1]
    if extension in ICON_EXTS:
        new_name = f.name + extension
        os.rename(f.name, new_name)
        return new_name
    return f.name
